% this routine is created by Ce Liu for batch SIFT flow matching 
% to find the correspondence from each of the test images and to its K-nearest
% neighbors from the training images
function batchSIFTflow(database,varargin)

% set the environmental parameters
% load the training and test data
setting = setEnvironment(database);
loadData;

%--------------------------------------------------------------------------------------------------------
% set the optional parameters
%--------------------------------------------------------------------------------------------------------
optParaDefault = {'alpha',setting.SIFTflowpara.alpha/255,...
                            'range',1:length(Dtest),...
                            'K',setting.K,...
                            'KNNmethod',setting.KNNmethod,...
                            'IsBackward',false,...
                            'IsOverwrite',true};
optPara = parsingVarArgIn(varargin,optParaDefault(1:2:end),optParaDefault(2:2:end));

% range is used for parallelization 
range         = optPara.range;
K                 = optPara.K;
KNNmethod   = optPara.KNNmethod;

% if IsBackward==true, then we also estimate the flow from the each of the
% nearest neighbors back to the input image
IsBackward = optPara.IsBackward;
% if IsOverwrite==true, then the flow data will be overwritten anyway; otherwise, the computation will skip on existin
% files
IsOverwrite = optPara.IsOverwrite;

% range has to be a row vector
if size(range,1)>size(range,2)
    range = range';
end

%--------------------------------------------------------------------------------------------------------
% inject the argument "alpha" to setting
%--------------------------------------------------------------------------------------------------------
alpha = optPara.alpha;
setting.SIFTflowpara.alpha = alpha*255;

%--------------------------------------------------------------------------------------------------------
%   create matching directory to save the flow data
%--------------------------------------------------------------------------------------------------------
MATCHINGPATH = fullfile(setting.path.matching.home,['alpha=' num2str(alpha)]);
if exist(MATCHINGPATH,'dir')~=7
    mkdir(MATCHINGPATH);
end


%--------------------------------------------------------------------------------------------------------
% set multi thread
%--------------------------------------------------------------------------------------------------------
if setting.IsMultiThread
    nCores = setMultiThread;
else
    nCores = 1;
end

%--------------------------------------------------------------------------------------------------------
% loop over the range
%--------------------------------------------------------------------------------------------------------
for m=range;
    if IsOverwrite == false && exist(fullfile(MATCHINGPATH,['matching' num2digits(m) '.mat']),'file') == 2
        fprintf('SIFT matching for test image %d exists!\n',m);
        continue;
    end
    fprintf('Find SIFT matching for test image %d ',m);
    
    %--------------------------------------------------------------------------------------------------------
    % retrieve nearest neighbors using a variety of methods
    %--------------------------------------------------------------------------------------------------------
    KNN = [];
    % use GIST L2 to retrieval K-nearest neighbors
    if strcmpi(KNNmethod,'gistL2') || strcmpi(KNNmethod,'all')
        knn = LMgistquery(gistTest(m,:), gistTraining,0);
        knn = knn(1:K);
        KNN = knn;
    end
    % use GIST normlized correlation to retrieval K-nearest neighbors
    if strcmpi(KNNmethod,'gistCorr') || strcmpi(KNNmethod,'all')
        knn = LMgistquery(gistTest(m,:), gistTraining,1);
        knn = knn(1:K);
        KNN = union(knn,KNN);
    end
    % use HOG histogram intersection to retrieve K-nearest neighbors
    if strcmpi(KNNmethod,'hog') || strcmpi(KNNmethod,'all')
        knn = LMhistintersectionquery(hogTest(m,:), hogTraining);
        knn = knn(1:K);
        KNN = union(knn,KNN);
    end    
    % use the ground-truth segmentation histogram intersection to retrieve K-NN
    if strcmpi(KNNmethod,'gt') || strcmpi(KNNmethod,'all')
        knn =  LMhistintersectionquery(objTest(m,:), objTraining);
        knn = knn(1:K);
        KNN = union(knn,KNN);
    end    
    clear knn;
    
    % load the image
    im1 = LMimread(Dtest,m,setting.path.images);
    
    % compute the dense SIFT feature of the query image
    sift1 = mexDenseSIFT(im1,setting.SIFTparam.cell_size,setting.SIFTparam.grid_spacing);
    
 % load the image
%     foo = Dtest(m).annotation;
%     im1 = imread(fullfile(fullfile(HOMEIMAGES,foo.folder),foo.filename));
%     im1 = cropImage(im1,SIFTparam.patch_size);
%     
%     subplottight(6,6,1);
%     imshow(showColorSIFT(sift1));
    fprintf('K=%d\n',length(KNN));
    Matching = cell(1,length(KNN));
    
    for j = 1:length(KNN)
        Matching{j}.im2 = LMimread(Dtraining,KNN(j),setting.path.images);
        if IsBackward
            [~,Matching{j}.seg,~] = LM2segments(Dtraining(KNN(j)),[],setting.path.images,setting.path.LMsegments);
        end
    end
    parfor j=1:length(KNN)
        
        % load the image
        
        
        % compute the dense SIFT feature of the nearest neighbor
        sift2 = mexDenseSIFT(Matching{j}.im2,setting.SIFTparam.cell_size,setting.SIFTparam.grid_spacing);
        
        % load the segment of the nearest neighbor
      
        % perform forward matching
        fprintf('\tMatching to No.%d NN (forward)...',j);
        tic;
        [vx,vy,energylist]=SIFTflowc2f(sift1,sift2,setting.SIFTflowpara);
        t=toc;
        fprintf('t=%.1f energy=%.2f done!\n',t/nCores,energylist(1).data(end)/511);
        Matching{j}.forward.vx=int16(vx);
        Matching{j}.forward.vy=int16(vy);
        Matching{j}.forward.energylist=energylist;
        Matching{j}.forward.objmin=energylist(1).data(end);

        % perform backward matching
        if IsBackward
            fprintf('\tMatching from No.%d NN (backward)...',j);
            tic;
            [vx,vy,energylist]=SIFTflowc2f(sift2,sift1,setting.SIFTflowpara,false,Matching(j).seg);
            t=toc;
            fprintf('t=%.1f energy=%.2f done!\n',t/nCores,energylist(1).data(end)/511);
            Matching{j}.backward.vx=int16(vx);
            Matching{j}.backward.vy=int16(vy);
            Matching{j}.backward.energylist=energylist;
            Matching{j}.backward.objmin=energylist(1).data(end);
        end        
        
%         foo = Dtraining(knn(j)).annotation;
%         im2 = im2double(imread(fullfile(fullfile(HOMEIMAGES,foo.folder),foo.filename)));
%         im2 = cropImage(im2,SIFTparam.patch_size);
%         warpI2 = warpImage(im2,matching(j).vx,matching(j).vy);
%         
%         [u,v]=ind2sub([6,5],j);
%         subplottight(6,6,sub2ind([6 6],v+1,u));
%         imshow(warpI2);drawnow
%         title(num2str(energylist(end)));
    end
    for j = 1:length(KNN)
        matching(j) = Matching{j};
    end
    average_energy=0;
    for j=1:length(KNN)
        average_energy=average_energy+matching(j).forward.objmin;
    end
    save(fullfile(MATCHINGPATH,['matching' num2digits(m) '.mat']),'matching','KNN');
    fprintf('\nAverage energy=%f\n\n',average_energy/(K*511));
end