% procedure to perform object recognition using SIFT flow
% only the best match is used to test the performance
% written by Ce Liu
function [recogRate,testArray]=objectRecognition(database,varargin)

% set the environmental parameters
% load the training and test data
setting = setEnvironment(database);
loadData;

%--------------------------------------------------------------------------------------------------------
% set the optional parameters
%--------------------------------------------------------------------------------------------------------
optParaDefault = {'KNNmethod',setting.KNNmethod,...
                            'nCandidates',10,...
                             'nNeighbors',50,...
                             'prior_weight',0.1,...
                             'spatial_weight',60,...
                             'trainingRatio',1,...
                             'isDisplay',true,...
                             'isPrint',true,...
                             'isSaveEps',true,...
                             'isOverwriteLikelihoodFile',false,...
                             'epsilon',10000,...
                             'range',1:length(Dtest),...
                             'alpha',setting.SIFTflowpara.alpha/255};
optPara = parsingVarArgIn(varargin,optParaDefault(1:2:end),optParaDefault(2:2:end));

%--------------------------------------------------------------------------------------------------------
% set the default parameters
%--------------------------------------------------------------------------------------------------------
nCandidates    = optPara.nCandidates;
nNeighbors      = optPara.nNeighbors;
prior_weight   = optPara.prior_weight;
spatial_weight = optPara.spatial_weight;
isDisplay           = optPara.isDisplay;
isPrint                 = optPara.isPrint;
isSaveEps         = optPara.isSaveEps;
alpha                  = optPara.alpha;
KNNmethod       = optPara.KNNmethod;
epsilon               = optPara.epsilon;
range                 = optPara.range;
trainingRatio    = optPara.trainingRatio;

% validate method
switch lower(KNNmethod)
    case lower('gistL2')
        KNNmethod = 'gistL2';
    case lower('gistCorr')
        KNNmethod = 'gistCorr';
    case lower('hog')
        KNNmethod = 'hog';
    case lower('obj')
        KNNmethod = 'obj';
    otherwise
        error(['The method ' KNNmethod ' is unknown!']);
end

K0 = nNeighbors;
K = K0;

%--------------------------------------------------------------------------------------------------------
% set the subfolder of matching (to load the matching data)
%--------------------------------------------------------------------------------------------------------
subfolder=['method=' KNNmethod ' alpha=' num2str(alpha)];
MATCHING = fullfile(setting.path.matching.home, ['alpha=' num2str(alpha)]);
if exist(MATCHING,'dir')~=7
    error(['The matching folder ' MATCHING ' does not exist!']);
end

%--------------------------------------------------------------------------------------------------------
% set the directories for displaying the results of object recognition
%--------------------------------------------------------------------------------------------------------
if isDisplay
    outputfolder=[subfolder ' nCandidates=' num2str(nCandidates) ' K=' num2str(K) ' epsilon=' num2str(epsilon) ...
                ' spatial=' num2str(spatial_weight) ' prior=' num2str(prior_weight)];
    RECOGJPGPATH=fullfile(setting.path.recog.jpg,KNNmethod,outputfolder);
    RECOGEPSPATH=fullfile(setting.path.recog.eps,KNNmethod,outputfolder);
    if exist(RECOGJPGPATH,'dir')~=7
        mkdir(RECOGJPGPATH);
    end
    if exist(RECOGEPSPATH,'dir')~=7
        mkdir(RECOGEPSPATH);
    end
end

%--------------------------------------------------------------------------------------------------------
% set the directory for the likelihood
%--------------------------------------------------------------------------------------------------------
LIKELIHOOD = fullfile(setting.path.likelihood,KNNmethod,[subfolder ' nCandidates=' num2str(nCandidates) ' K=' num2str(K)  ' epsilon=' num2str(epsilon)]);
if exist(LIKELIHOOD,'dir')~=7
    mkdir(LIKELIHOOD);
end

objprior = [];
load(fullfile(setting.path.database,'objectinfo.mat')); % load visualization information
load(fullfile(setting.path.database,'objectPrior.mat')); %load matrix "objprior" and "names"
nClasses=length(names);

% initialize counting matrices
matrClassified=zeros(nClasses);
vectTotal=zeros(nClasses,1);

totalCorrect=0;
totalPixels=0;

if isDisplay
    figure;
end

% sometimes we don't want all the training data
if trainingRatio<1
    nTraining = length(Dtraining);
    trainingIdx = randperm(nTraining);
    trainingIdx = trainingIdx(1:round(nTraining*trainingRatio));
end

maxK = 120;

for m = range
 %for m=1:length(Dtest)
% for m =1:200
%range = [1:28, 101:128];
%for m=range
%for m=54:54

    % load the ground-truth segmentation
    [~,seg_truth]=LM2segments(Dtest(m),[],setting.path.images,setting.path.LMsegments);
    
    K = K0;
    clear idx ik knn KNN energylist;
    switch KNNmethod
        case 'gistL2'
            [knn,dist] = LMgistquery(gistTest(m,:), gistTraining);
        case 'gistCorr'
            [knn,dist] = LMgistquery(gistTest(m,:), gistTraining,1);
        case 'hog'
            [knn,dist] = LMhistintersectionquery(hogTest(m,:), hogTraining);
        case'obj'            
            [knn,dist] =  LMhistintersectionquery(objTest(m,:), objTraining);
        otherwise
    end

    if trainingRatio<1
        dist = dist(1:maxK);
        knn = knn(1:maxK);
        [knn,~,idx] = intersect(trainingIdx,knn);
        if isempty(knn)  
            totalPixels = totalPixels + length(seg_truth(:)>0);
            if isPrint
                    fprintf('No. %d  K=%d Recognition rate: %f   total: %f\n',m,0,0,totalCorrect/totalPixels);
            end                    
            continue;
        end
        dist = dist(idx);
        [dist,idx] = sort(dist,'ascend');
        knn = knn(idx);
        if length(knn)>K
            dist = dist(1:K);
        end
        K = max(1,sum(dist<=dist(1)*epsilon));
        knn = knn(1:K);
        dist = dist(1:K);
    else
        dist = dist(1:K);
        K = max(1,sum(dist<=dist(1)*epsilon));
        knn = knn(1:K);
        dist = dist(1:K);
    end
    
    
    % load the image
    foo = Dtest(m).annotation;
    im1 = LMimread(Dtest,m,setting.path.images);
    
    % load the matching
    if exist(fullfile(MATCHING,['matching' num2digits(m) '.mat']),'file')~=2
        continue;
    end
    load(fullfile(MATCHING,['matching' num2digits(m) '.mat']));
    
    % find the intersection with KNN
    % ik is where knn is in KNN
    [knn,ik] = intersect(KNN,knn);
    if length(knn)~=K
        error('Cannot find nearest neighbors in the precomputed union!');
    end
    
    for j=1:K
        energylist(j)=matching(ik(j)).forward.energylist(1).data(end);
    end
    [~,idx]=sort(energylist);
    
    

    likelihoodfile = fullfile(LIKELIHOOD,['likelihood_' num2digits(m) '.mat']);
    
    % check whether the likelihood file exists or not
    % if does not exist, then create likelihood
    if exist(likelihoodfile,'file')==2 && optPara.isOverwriteLikelihoodFile == false
        load(likelihoodfile);
    else
        % compute the dense SIFT feature of the query image
        sift1 = mexDenseSIFT(im1,setting.SIFTparam.cell_size,setting.SIFTparam.grid_spacing);
        sift1 = im2double(sift1);
        
        % count the valid object classes from the nearest neighbors
        histogram = zeros(1,nClasses);
        
%         if max(idx) > K
%             error('Something wrong');
%         end
        
        for j = 1:min(nCandidates,K)
            
            index = knn(idx(j));
            [~,seg]=LM2segments(Dtraining(index),[],setting.path.images,setting.path.LMsegments);
            
            seg = double(seg(:));
            N = max(seg);
            foo = hist(seg,N+1);
            foo(N+2:nClasses)=0;
            histogram = histogram + foo;
        end
        histogram(1) = max(histogram(1),1); % we always include "unlabled" in the first place
        if nClasses < 100
            histogram = histogram + 1;
        end
        ValidClassIdx = find(histogram>0);
        nValidClasses = length(ValidClassIdx);
        LookupTable = zeros(1,nClasses+1);
        LookupTable(ValidClassIdx) = 1:nValidClasses;
        
        likelihood=ones(256,256,nValidClasses)*10;

        % load the segmentation of the best matches
        for j=1:min(nCandidates,K);
            %fprintf('\t\t Get information from candidate %d...',j);
            index=knn(idx(j));
            [~,seg]=LM2segments(Dtraining(index),[],setting.path.images,setting.path.LMsegments);
            
            % convert seg to the valid segmentation index using the lookup table
            seg = uint16(interp1(0:nClasses,LookupTable,double(seg),'nearest'));
            
            % load the image of the nearest neighbor
            im2 = LMimread(Dtraining,index,setting.path.images);
            
            % compute the dense SIFT feature of the nearest neighbor
            sift2 = mexDenseSIFT(im2,setting.SIFTparam.cell_size,setting.SIFTparam.grid_spacing);
            sift2 = im2double(sift2);

            % warp the segmenation of the best match to the query
            matching(ik(idx(j))).forward.vx = double(matching(ik(idx(j))).forward.vx);
            matching(ik(idx(j))).forward.vy = double(matching(ik(idx(j))).forward.vy);
            warpSeg   = int16(warpSegmentation(seg,matching(ik(idx(j))).forward.vx,matching(ik(idx(j))).forward.vy));
            [warpSIFT,mask]  = mexWarpImageInt(sift2,matching(ik(idx(j))).forward.vx,matching(ik(idx(j))).forward.vy);

            %warpSeg = int16(seg);
    
            % compute the matching error
            matchingError=sum(abs(warpSIFT-sift1),3);
            matchingError(mask==0)=0;

            % loop over each class
            for i=1:nValidClasses
                index=find(warpSeg==i);
                if ~isempty(index)
                    foo=likelihood(:,:,i);
                    foo(index)=min(foo(index),matchingError(index));
                    likelihood(:,:,i)=foo;
                end
            end
            %fprintf('done!\n');
        end
        save(likelihoodfile,'likelihood','histogram','ValidClassIdx','nValidClasses');
    end
    
    weight=imsmooth(im1,0.4);
    
    if exist('ValidClassIdx','var') == 1
        ObjPrior = zeros(256,256,nValidClasses);
        for i =1:nValidClasses
            ObjPrior(:,:,i) = objprior(:,:,ValidClassIdx(i));
        end
        tic;
        [warpSegMAP,energylist]=mexIntegrateAnnotation(likelihood(:,:,2:end)+ObjPrior(:,:,2:end)*prior_weight,[spatial_weight,10,3],weight*spatial_weight);
        t=toc;
        warpSeg=warpSegMAP+1;
        if nValidClasses >1
            warpSeg = int16(interp1(0:nValidClasses-1,ValidClassIdx-1,warpSeg,'nearest'));
        else
            warpSeg = int16(ones(size(warpSeg))*(ValidClassIdx-1));
        end
    else
        tic;
        [warpSegMAP,energylist]=mexIntegrateAnnotation(likelihood(:,:,2:end)+objprior(:,:,2:end)*prior_weight,[spatial_weight,10,3],weight*spatial_weight);
        t=toc;
        warpSeg=int16(warpSegMAP+1);
    end
    
    % count the correct matches and the total number of pixels
    foo = (seg_truth==warpSeg) & (seg_truth~=0);
    nCorrect = sum(foo(:));
    totalCorrect=totalCorrect+nCorrect;
    foo = (seg_truth~=0);
    nTotal = sum(foo(:));
    totalPixels=totalPixels+nTotal;
    testArray(m).nPixels = nTotal;
    testArray(m).nCorrect =nCorrect;
    testArray(m).rate = nCorrect/nTotal;
    testArray(m).energy = energylist(end);
    testArray(m).dist = dist;
    if isPrint
        fprintf('No. %d  K=%d Recognition rate: %f   total: %f\n',m,K,nCorrect/nTotal,totalCorrect/totalPixels);
    end
    
    % do the matching for each category
    foo=histc(seg_truth(:),0:nClasses);
    %idx=find(foo>0);
    existence=zeros(nClasses,1);
    for i=1:nClasses
        mask=(seg_truth==i-1);
        mask2=(warpSeg==i-1);
        if sum(mask(:))+sum(mask2(:))>0
            existence(i)=1;
        end
        vectTotal(i)=vectTotal(i)+sum(mask(:));
        foo=int16(mask).*warpSeg;
        foo=foo(mask>0);
        if ~isempty(foo)
            for j=1:nClasses
                matrClassified(i,j)=matrClassified(i,j)+sum(foo==j-1);
            end
        end
    end
    
    if isDisplay
        clf;
        subplottight(1,7,1);imshow(im1);
        
        foo = Dtraining(knn(idx(1))).annotation;
        
        index=knn(idx(1));
        % load the image of the nearest neighbor
        im2 = LMimread(Dtraining,index,setting.path.images);        
        
        subplottight(1,7,2);imshow(im2);
        
        %[foo,seg]=LM2segments(Dtraining(knn(idx(1))),[],HOMEIMAGES,HOMELMSEGMENTS);
        [~,seg]=LM2segments(Dtraining(knn(idx(1))),[],setting.path.images,setting.path.LMsegments);
        
        %seg       = cropImage(seg,SIFTparam.patch_size);
        subplottight(1,7,3);imshow(seg2color(seg,objectinfo.objcolormap));
        
        vx=double(matching(ik(idx(1))).forward.vx);
        vy=double(matching(ik(idx(1))).forward.vy);
        warpI2 = warpImage(im2,vx,vy)/255;
        subplottight(1,7,4);imshow(warpI2);
        
        subplottight(1,7,5);imshow(seg2color(warpSeg,objectinfo.objcolormap));
        subplottight(1,7,6);imshow(seg2color(seg_truth,objectinfo.objcolormap));
        
        subplottight(1,7,7);showObjColorMapIndex(existence,objectinfo.names,objectinfo.objcolormap);
        
        drawnow;

        print('-djpeg', '-r100', fullfile(RECOGJPGPATH,['matching' num2digits(m) ' rate=' num2str(nCorrect/nTotal) '.jpg']));
        saveas(gcf,fullfile(RECOGEPSPATH,['matching' num2digits(m) '.eps']),'psc2');
    end    
end

% sort the index
[~,idx]=sort(objectinfo.counts,'descend');
for i=1:nClasses
    H(i,:)=matrClassified(i,idx);
end
H=H(idx,:);

W=repmat(vectTotal(idx),[1 nClasses]);

M=H./W;



performance=diag(M);
%performance=performance(1:end-3);

%--------------------------------------------------------------------------------------------------------
% treat unlabeled
%--------------------------------------------------------------------------------------------------------
for i=1:length(idx)
    if strcmp(names{idx(i)},'unlabeled')==1
        hh=i;
        break;
    end
end
idx=[idx(1:hh-1);idx(hh+1:end)];
performance=[performance(1:hh-1);performance(hh+1:end)];
Names=objectinfo.names(idx);
Colormap = objectinfo.objcolormap(idx,:);
recogRate=totalCorrect/totalPixels;

filename =  ['ConfusionMatrix method=' KNNmethod ' nCandidates=' num2str(nCandidates) ' K=' num2str(K)  ' epsilon=' num2str(epsilon)...
                ' spatial=' num2str(spatial_weight) ' prior=' num2str(prior_weight) '.mat'];
save(fullfile(setting.path.recog.home,filename),'M','performance','recogRate');

% % figure;h=bar(performance+0.001);
% % colormap(objectinfo.objcolormap(idx,:));
% % 
% % ch=get(h,'Children');
% % fvd=get(ch,'Faces');
% % fvcd=get(ch,'FaceVertexCData');
% % for i=1:length(idx)
% %     fvcd(fvd(i,:))=i;
% % end
% % set(ch,'FaceVertexCData',fvcd);
% % 
% % set(gca,'xtick',[]);
% % Names=names(idx);
% % for i=1:length(idx)
% %     text(i,-0.01,Names{i},'Rotation',270);
% % end
% % ylim([0,1]);
% % xlim([0,31]);
% % 
% % title(['Recognition rate: ' num2str(recogRate*100) '%'],'FontSize',20);
if isSaveEps
    if strcmp(database,'datasetLMO') % if database = dataset1 then ignore the last three categories
        dispObjRecogRate(database,performance(1:end-3),recogRate,Names(1:end-3),Colormap(1:end-3,:));
    else
        dispObjRecogRate(database,performance,recogRate,Names,Colormap);
    end
    saveas(gcf,fullfile(setting.path.recog.eps,[subfolder ' nCandidates=' num2str(nCandidates) ' K=' num2str(K)  ' epsilon=' num2str(epsilon)...
                               ' spatial=' num2str(spatial_weight) ' prior=' num2str(prior_weight)...
                               ' RecogRate=' num2str(recogRate) '.eps']),'psc2');
end